package com.codigoparallevar.minicards.parts.buttons;

import android.graphics.Color;
import android.graphics.Paint;
import android.util.Log;

import com.codigoparallevar.minicards.PartInstantiator;
import com.codigoparallevar.minicards.ScrolledCanvas;
import com.codigoparallevar.minicards.types.PartConnection;
import com.codigoparallevar.minicards.types.Tuple2;
import com.codigoparallevar.minicards.types.connectors.input.InputConnector;
import com.codigoparallevar.minicards.types.Moveable;
import com.codigoparallevar.minicards.types.connectors.input.SignalInputConnector;
import com.codigoparallevar.minicards.types.connectors.output.OutputConnector;
import com.codigoparallevar.minicards.types.Part;
import com.codigoparallevar.minicards.types.PartGrid;
import com.codigoparallevar.minicards.types.RoundInputConnector;
import com.codigoparallevar.minicards.types.RoundOutputConnector;
import com.codigoparallevar.minicards.utils.Serializations;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class RoundButton implements Part {
    private final String _id;
    private final PartGrid _partGrid;
    private int _xCenter;
    private int _yCenter;
    private final int _innerRadius;
    private final int _outerRadius;
    private final int _outerRadiusThickness = 10;
    private final int _pathRunWay = 200;
    private List<OutputConnector> _outputConnectors;
    private final RoundOutputConnector _pressedOuputConnector;
    private final static int DEFAULT_INNER_RADIUS = 80;
    private final static int DEFAULT_OUTER_RADIUS = 100;


    private RoundButton(String id, PartGrid partGrid, int xCenter, int yCenter, int innerRadius, int outerRadius) {
        _id = id;

        _partGrid = partGrid;
        _xCenter = xCenter;
        _yCenter = yCenter;
        _innerRadius = innerRadius;
        _outerRadius = outerRadius;

        // Create connectors
        _pressedOuputConnector = new RoundOutputConnector(
                this,
                _partGrid,
                getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius());

        _outputConnectors = new LinkedList<>();
        _outputConnectors.add(_pressedOuputConnector);
    }

    public RoundButton(PartGrid partGrid, int xCenter, int yCenter, int innerRadius, int outerRadius) {
        this(UUID.randomUUID().toString(), partGrid, xCenter, yCenter, innerRadius, outerRadius);
    }

    @Override
    public int getLeft() {
        return _xCenter - _outerRadius / 2;
    }

    @Override
    public int getRight() {
        return _xCenter + _outerRadius / 2;
    }

    @Override
    public int getTop() {
        return _yCenter - _outerRadius / 2;
    }

    @Override
    public int getBottom() {
        return _yCenter + _outerRadius / 2;
    }

    @Override
    public void draw(ScrolledCanvas canvas, boolean devMode) {
        Paint foregroundPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        foregroundPaint.setColor(Color.WHITE);

        Paint backgroundPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        backgroundPaint.setColor(Color.BLACK);

        if (devMode){
            drawConnector(canvas);
            drawWires(canvas, devMode);
        }

        canvas.drawCircle(_xCenter, _yCenter, _outerRadius, foregroundPaint);
        canvas.drawCircle(_xCenter, _yCenter, _outerRadius - _outerRadiusThickness, backgroundPaint);
        canvas.drawCircle(_xCenter, _yCenter, _innerRadius, foregroundPaint);
    }

    private void drawConnector(ScrolledCanvas canvas) {
        Paint connectorPaint = new Paint(Paint.ANTI_ALIAS_FLAG);
        connectorPaint.setColor(Color.RED);

        canvas.drawCircle(getOutputConnectorCenterX(), getOutputConnectorCenterY(),
                getOutputConnectRadius(),
                connectorPaint);
    }

    private int getOutputConnectorCenterX() {
        return _xCenter + _outerRadius;
    }

    private int getOutputConnectorCenterY() {
        return _yCenter;
    }

    private int getOutputConnectRadius() {
        return _innerRadius;
    }

    private void drawWires(ScrolledCanvas canvas, boolean devMode) {
        for (OutputConnector outputConnector : _outputConnectors){
            outputConnector.drawWires(canvas, devMode);
        }
    }

    @Override
    public void moveEnd(int x, int y) {
        _xCenter = x;
        _yCenter = y;

        // Move connectors too
        _pressedOuputConnector.updatePosition(
                getOutputConnectorCenterX(),
                getOutputConnectorCenterY());
    }

    @Override
    public void drop(int x, int y) {
        moveEnd(x, y);
    }

    @Override
    public void touched() {
        Log.d("RoundButton", "Round button touched");

        _pressedOuputConnector.sendSignal();
    }

    @Override
    public List<SignalInputConnector> getSignalInputConnectors() {
        return Collections.emptyList();
    }

    @Override
    public List<OutputConnector> getOutputConnectors() {
        return _outputConnectors;
    }

    @Override
    public JSONObject serialize() throws JSONException {
        JSONObject serialized = new JSONObject();

        serialized.put("id", _id);
        serialized.put("x_center", _xCenter);
        serialized.put("y_center", _yCenter);
        serialized.put("inner_radius", _innerRadius);
        serialized.put("outer_radius", _outerRadius);
        serialized.put("outer_radius_thickness", _outerRadiusThickness);
        serialized.put("on_pressed_output_connector",
                Serializations.serialize(serializeConnectionEndpoints()));

        return serialized;
    }

    private List<Map<String, String>> serializeConnectionEndpoints() {
        List<Map<String, String>> serializedData = new LinkedList<>();

        for (Tuple2<String, String> endpoint : _pressedOuputConnector.getConnectionEndpoints()){
            serializedData.add(PartConnection.serialize(endpoint.item1, endpoint.item2));
        }

        return serializedData;
    }

    @Override
    public void sendSignal(RoundInputConnector roundInputConnector) {
        // @TODO: REMOVE THE NEED FOR THIS
    }

    @Override
    public String getId() {
        return _id;
    }

    @Override
    public InputConnector getConnectorWithId(String inputConnectorId) {
        return null;
    }

    @Override
    public String getConnectorId(InputConnector inputConnector) {
        return null;
    }

    public static Tuple2<Part, List<PartConnection>> deserialize(PartGrid partGrid, JSONObject data) throws JSONException {
        String id = data.getString("id");
        int xCenter = data.getInt("x_center");
        int yCenter = data.getInt("y_center");

        int innerRadius = data.getInt("inner_radius");
        int outerRadius = data.getInt("outer_radius");

        RoundButton button = new RoundButton(id, partGrid, xCenter, yCenter, innerRadius, outerRadius);

        List<PartConnection> connections = new LinkedList<>();

        JSONArray connectorOuts = data.getJSONArray("on_pressed_output_connector");
        for (int i = 0; i < connectorOuts.length(); i++){
            connections.add(PartConnection.deserialize(
                    button._pressedOuputConnector,
                    connectorOuts.getJSONObject(i)));
        }

        return new Tuple2<Part, List<PartConnection>>(button, connections);
    }

    @Override
    public boolean containsPoint(int x, int y) {
        return ((Math.abs(x - _xCenter) <= _outerRadius)
                && (Math.abs(y - _yCenter) <= _outerRadius));
    }

    @Override
    public Moveable getMoveable() {
        return this;
    }


    @Override
    public void unlink() {
        for (InputConnector input : getSignalInputConnectors()) {
            input.unlink();
        }
    }

    public static PartInstantiator getInstantiator() {
        return new PartInstantiator() {
            @Override
            protected Part instantiate(PartGrid grid, Tuple2<Integer, Integer> center) {
                return new RoundButton(grid,
                        center.item1, center.item2,
                        DEFAULT_INNER_RADIUS, DEFAULT_OUTER_RADIUS);
            }
        };
    }
}
